from typing import Tuple
import logging

from tritonclient.grpc import InferResult # Purely for typing
from tritonclient import grpc
from tritonclient import http

import torch

logger = logging.getLogger(__name__)


class FastNNClient:
    """FastNN client class for pushing requests to the Triton Inference Server


    Usage:
    ```python
    >>> client = FastNNClient(url="127.0.0.1:8000", model_name="distilbert-squad", model_version="1")
    >>> client.request(batch=batch)
    ```

    **Parameters:**

    * **url** - String url of Triton Inference Server. Defaults to 127.0.0.1:8000
    * **model_name** - String name of model in `model_repository` directory
    * **model_version** - String model version name
    * **client_type** - A string for choosing between http and grpc protocol ["grpc", "http"] (Defaults to "grpc")
    * **verbose** - Bool for log verbosity
    """

    def __init__(
        self,
        url: str = "127.0.0.1:8000",
        model_name: str = "distilbert-squad",
        model_version: str = "1",
        client_type: str = "grpc",
        verbose: bool = False,
    ):

        if client_type=="grpc":
            self.client_pkg = grpc
        elif client_type=="http":
            self.client_pkg = http
        else:
            ValueError("Paramater 'client_type' must be either grpc or http")

        self.url = url
        self.model_name = model_name
        self.model_version = model_version
        self.client_type = client_type

        self.triton_client = self.client_pkg.InferenceServerClient(url=url, verbose=verbose)
        self.model_metadata = self.triton_client.get_model_metadata(
            model_name=model_name, model_version=model_version
        )
        self.model_config = self.triton_client.get_model_config(
            model_name=model_name, model_version=model_version
        )

    def request(
        self, batch: Tuple[torch.Tensor], binary_data: bool = False
    ) -> InferResult:
        """Runs a request with the `batch` input that can be generated from a FastNN `Processor`

        * **batch** - Tuple of torch tensors, typically batch inputs from a dataloader
        """
        if self.client_type=="grpc":
            return self.request_grpc(batch=batch)
        elif self.client_type=="http":
            return self.request_http(batch=batch)
        else:
            ValueError("Paramater 'client_type' must be either grpc or http")

    def request_http(
        self, batch: Tuple[torch.Tensor], binary_data: bool = False
    ) -> InferResult:
        """Runs an http request with the `batch` input that can be generated from a FastNN `Processor`

        * **batch** - Tuple of torch tensors, typically batch inputs from a dataloader
        """
        inputs_metadata = self.model_metadata["inputs"]
        outputs_metadata = self.model_metadata["outputs"]

        # Assert batch input matches triton model metadata
        assert len(batch) == len(inputs_metadata)

        inputs = []
        for i, metadata in enumerate(inputs_metadata):
            inp = self.client_pkg.InferInput(
                metadata["name"], tuple(batch[i].shape), metadata["datatype"]
            )
            inp.set_data_from_numpy(batch[i].cpu().numpy(), binary_data=binary_data)
            inputs.append(inp)

        outputs = []
        for i, metadata in enumerate(outputs_metadata):
            out = self.client_pkg.InferRequestedOutput(metadata["name"], binary_data=binary_data)
            outputs.append(out)

        response = self.triton_client.infer(
            model_name=self.model_name,
            model_version=self.model_version,
            inputs=inputs,
            outputs=outputs,
        )

        return response

    def request_grpc(
        self, batch: Tuple[torch.Tensor], binary_data: bool = False
    ) -> InferResult:
        """Runs a grpc request with the `batch` input that can be generated from a FastNN `Processor`

        * **batch** - Tuple of torch tensors, typically batch inputs from a dataloader
        """
        inputs_metadata = self.model_metadata.inputs
        outputs_metadata = self.model_metadata.outputs

        # Assert batch input matches triton model metadata
        assert len(batch) == len(inputs_metadata)

        inputs = []
        for i, metadata in enumerate(inputs_metadata):
            inp = self.client_pkg.InferInput(
                metadata.name, tuple(batch[i].shape), metadata.datatype
            )
            inp.set_data_from_numpy(batch[i].cpu().numpy())
            inputs.append(inp)

        outputs = []
        for i, metadata in enumerate(outputs_metadata):
            out = self.client_pkg.InferRequestedOutput(metadata.name)
            outputs.append(out)

        response = self.triton_client.infer(
            model_name=self.model_name,
            model_version=self.model_version,
            inputs=inputs,
            outputs=outputs,
        )

        return response


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fastnn',
 'fastnn.nn',
 'fastnn.processors',
 'fastnn.processors.cv',
 'fastnn.processors.nlp',
 'fastnn.utils',
 'fastnn.utils.qa']

package_data = \
{'': ['*']}

install_requires = \
['transformers>=4.0.0,<5.0.0',
 'tritonclient[all]>=2.18.0,<3.0.0',
 'wget>=3.2,<4.0']

extras_require = \
{'docs': ['mkdocs>=1.1.2,<2.0.0',
          'mkdocs-material>=6.1.5,<7.0.0',
          'mkdocstrings>=0.18.0,<0.19.0'],
 'ios': ['coremltools>=4.0,<5.0'],
 'jupyter': ['jupyter>=1.0.0,<2.0.0',
             'jupyterlab>=2.2.9,<3.0.0',
             'matplotlib>=3.3.3,<4.0.0'],
 'torch': ['torch>=1.0.0,<2.0.0', 'torchvision<1.0.0']}

setup_kwargs = {
    'name': 'fastnn',
    'version': '0.2.0',
    'description': 'A python library and framework for fast neural network computations.',
    'long_description': '# Fast Neural Networks (FastNN)\n\nA framework for deploying serializable and optimizable neural net models at scale in production via. the NVIDIA Triton Inference Server.\n\n<p align="center">\n    <a href="https://hub.docker.com/r/aychang/fastnn">\n        <img src="https://img.shields.io/docker/cloud/build/aychang/fastnn"\n    </a>\n    <a href="https://badge.fury.io/py/fastnn">\n        <img src="https://badge.fury.io/py/fastnn.svg">\n    </a>\n    <a href="https://github.com/aychang95/fastnn/blob/master/LICENSE">\n        <img src="https://img.shields.io/github/license/aychang95/fastnn">\n    </a>\n</p>\n\n## [**FastNN Docker Release Selector (Ubuntu 18.04)**](https://andrewchang.dev/fastnn/index.html#fastnn-docker-release-selector-ubuntu-1804)\n\n## [Documentation](https://andrewchang.dev/fastnn) - Guides, Models, API References\n\n## Features:\n  - **Data Processing**\n    - Intuitive data processing modules for encoding human-readible data into tensors compatible with deep learning libraries\n  - **Model Exporting**\n    - FastNN torch modules and tools for exporting models via. `TorchScript` tracing and scripting to a production environment\n  - **Model Zoo**\n    - Various exported models hosted in this repo via. git-lfs and AWS S3. Includes models from the HuggingFace\'s Transformers and \n    TorchVision\n  - **Model Deployment**\n    - Deploy models using Triton Inference Server on CPU/GPU-compatible server(s) with helm or docker\n  - **FastNN Client**\n    - Client wrapper for Triton Inference Server\'s client module for programmatic requests with python\n\n\n## Pre-Requisites:\n\nGit LFS is required if you\'d like to use any of the models provided by FastNN in `./model_repository`.\n\nCloning this repository without Git LFS will clone a repository with LFS pointers, not the actual model.\n\nAfter the repository is cloned and Git LFS is installed, set up with `git lfs install`.\n\nDownload specific models with:\n\n```sh\ngit lfs pull --include="model_repository/<path-to-model-dir>" --exclude=""\n```\n\nDownload ALL models with:\n\n```sh\ngit lfs pull\n\n# Or\n#git lfs pull --include="*" --exclude=""\n```\n\n\n## Quickstart and Installation:\n\n### *Pre-requisites:*\n\nRequirements: Python 3.7+, PyTorch 1+, TorchVision 0.7+, Triton Client\n\nOptional: CUDA Compatible GPU, NVIDIA Drivers, cudnn (PyTorch pre-built wheels)\n\n1. To install PyTorch with TorchVision, please refer to the installation instructions on their web page [here](https://pytorch.org/get-started/locally/#start-locally).\n\n2. The tritonclient package wheels are not hosted on the public PyPI server. We need to add the address of NVIDA\'s private python package index to the environment. You can complete these steps and install the tritonclient package by running the following.\n\n```sh\n# If you cloned this repo, you can just uncomment and run the one line below\n#sh ./scripts/install_triton_client.\npip install nvidia-pyindex\npip install tritonclient[all]\n```\n\n### **Install via. pip**\n\nOnce the above requirements are set, you can install fastnn with the command below:\n\n```sh\npip install fastnn\n```\n\nIf you are comfortable with the latest default stable releases of PyTorch you can skip step 1 in the pre-requisites and run `pip install fastnn[torch]` instead.\n\n\n### **Install from source with Poetry for development**\n\nYou will need to install poetry by referring to the installation instructions on their web page [here](https://python-poetry.org/docs/#installation).\n\nAfter cloning the repository, just run `poetry install` to install with the .lock file.\n\nActivate the virtual environment with the command below:\n\n```sh\npoetry shell\n```\n\n\n### **Docker**\n\nOfficial FastNN images are hosted on [Docker Hub](https://hub.docker.com/r/aychang/fastnn).\n\nSelect FastNN package and image versions by referencing the [documentation](https://andrewchang.dev/fastnn/index.html#fastnn-docker-release-selector-ubuntu-1804). Development and runtime environments are available.\n\nJupyter lab and notebook servers are accessible with notebook examples and terminal access `localhost:8888` with every image.\n\n\nRun the latest FastNN image by running below:\n\n```sh\ndocker run --gpus all --rm -it -p 8888:8888 aychang/fastnn:latest\n```\n\nRun images with specific configurations as can see in the example command below:\n\n```sh\ndocker run --gpus all --rm -it -p 8888:8888 aychang/fastnn:0.1.0-cuda11.0-runtime-ubuntu18.04-py3.7\n\n```\n\n\n## Triton Inference Server: Local or Kubernetes Cluster\n\n\n### **Local Deployment**\n\nRequirements:\n  - Docker 19.03+\n\nGPU Inference Requirements:\n\n  - NVIDIA CUDA-Compatible GPU\n  \n  - [NVIDIA Container Toolkit (nvidia-docker)](https://docs.nvidia.com/datacenter/cloud-native/container-toolkit/install-guide.html)\n\nLocal deployment of the Triton Server uses the EXPLICIT model control mode. Local models must be explicitly specified with the `--load-model` \nargument in `./scripts/start_triton_local.sh`\n\n```sh\n\nexport MODEL_REPOSITORY=$(pwd)/model_repository\nsh ./scripts/start_triton_local.sh\n\n```\n\n\n### **Helm Chart install in Kubernetes Cluster**\n\nRequirements: kubectl 1.17+, Helm 3+, Kubernetes 1.17+\n\nYou can currently install the local FastNN helm chart with the following instuctions:\n\n```sh\n\ncd ./k8s\nhelm install fastnn .\nexport MODEL_REPOSITORY=$(pwd)/model_repository\n\n```\n\nNote: The current local helm chart installation deploys Triton using the NONE model control mode. All models available in the S3 Model Zoo will be deployed...good luck. \nDeployed models and model control mode can be edited in the helm chart deployment configuration file.\n\n# License\n\nThis project is licensed under the terms of the MIT license.\n',
    'author': 'Andrew Chang',
    'author_email': 'aychang995@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://andrewchang.dev/fastnn/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from typing import List, Tuple
from collections import OrderedDict

import torch
from torch.utils.data import DataLoader, Dataset, TensorDataset
from torch.utils.data.dataloader import default_collate

from transformers import (
    AutoTokenizer,
)

from transformers.data.processors.squad import SquadFeatures, SquadResult

from fastnn.utils.qa.squad_metrics import (
    compute_predictions_logits,
)

from fastnn.processors.base_processing import Processor


logger = logging.getLogger(__name__)


class TransformersTokenTaggingProcessor(Processor):
    """Token Tagging Data Processor. Use this class to generate tensor inputs from human legible text/string data.
    This class can be used with a majority of the Bert architecture transformer models with a token-level predictive head
    for token classification from Hugging Face.

    Usage:
    ```python
    >>> processor = TransformersTokenTaggingProcessor(model_name_or_path="dbmdz/bert-large-cased-finetuned-conll03-english")
    >>> processor.process(text=["string"])

    **Parameters:**

    * **model_name_or_path** - String defining HF token tagging model/tokenizer's name
    * **label_strings** - List of strings that specify label strings with index as key for this specific processor
    ```
    """

    def __init__(
        self, model_name_or_path: str = "dbmdz/bert-large-cased-finetuned-conll03-english", label_strings: List[str]=[
                "O",       # Outside of a named entity
                "B-MISC",  # Beginning of a miscellaneous entity right after another miscellaneous entity
                "I-MISC",  # Miscellaneous entity
                "B-PER",   # Beginning of a person's name right after another person's name
                "I-PER",   # Person's name
                "B-ORG",   # Beginning of an organisation right after another organisation
                "I-ORG",   # Organisation
                "B-LOC",   # Beginning of a location right after another location
                "I-LOC"    # Location
            ]
    ):
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            use_fast=True
        )
        self.label_strings = label_strings

    def process(
        self,
        text: List[str],
        max_seq_length: int = 512,
    ) -> Tuple:
        """Generate torch `Dataset` object from query/context string pairs using specified tokenizer from HF.
        This provides clear tensor input representations for compatible models.

        Returns a `Dataset`

        * **text** - List of text strings
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        """

        tokens = self.tokenizer(text=text, return_tensors="pt", padding="max_length", truncation=True)
        dataset = TensorDataset(tokens["input_ids"], tokens["attention_mask"], )


        return dataset

    def process_batch(
        self,
        text: List,
        mini_batch_size: int = 8,
        max_seq_length: int = 512,
        use_gpu: bool = False,
    ) -> DataLoader:
        """Generate torch `DataLoader` object from text strings using specified tokenizer from HF.
        This provides clear tensor input representations for compatible models in an easy to use batch

        Returns a `DataLoader`

        * **text** - List of text strings
        * **mini_batch_size** - Batch size for inference
        * **max_seq_length** - Maximum context token length. Check model configs to see max sequence length the model was trained with
        * **use_gpu** - Bool for using gpu or cpu. If set True but no gpu devices available, model will default to using cpu
        """

        if use_gpu:
            if torch.cuda.is_available():
                device = torch.device("cuda")
            else:
                logger.info("GPU not available")
                device = torch.device("cpu")
        else:
            device = torch.device("cpu")

        dataset = self.process(
            text=text
        )

        dataloader = DataLoader(
            dataset,
            batch_size=mini_batch_size,
            collate_fn=lambda x: [t.to(device) for t in self._collate_fn(x)],
        )


        return dataloader

    def process_output(
        self,
        outputs: List[Tuple[torch.Tensor, torch.Tensor, torch.Tensor]],
    ):
        pass

    def process_output_batch(
        self,
        outputs: List
    ) -> List[List[Tuple[str, str]]]:
        """ Process output of Transformers NER model

        * **outputs** - List of batch output tensors from a model's forward pass
        """
        results = []
        for logits, input_ids in outputs:
            tokens_batch = [self.tokenizer.convert_ids_to_tokens(i) for i in input_ids]
            argmax_batch = [torch.argmax(o, dim=1) for o in logits]
            for i in range(len(tokens_batch)):
                # Filter out padding
                results.append([(token, self.label_strings[prediction]) for token, prediction in zip(tokens_batch[i], argmax_batch[i].cpu().numpy()) if token!='[PAD]'])
        return results

    def _collate_fn(self, data):
        batch = default_collate(data)
        return batch


from typing import Tuple

import torch
from transformers import AutoModelForTokenClassification


class NERModule(torch.nn.Module):
    """Module for Token Classification models from Transformers

    Usage:
    ```python
    >>> model = NERModule(model_name_or_path="dbmdz/bert-large-cased-finetuned-conll03-english")
    >>> output = model(*batch)
    ```

    **Parameters:**

    * **model_name_or_path** - String key of Transformer's pre-trained model hosted in Hugging Face's Model Repository
    """

    def __init__(
        self, model_name_or_path: str = "dbmdz/bert-large-cased-finetuned-conll03-english"
    ):
        super(NERModule, self).__init__()
        self.model = AutoModelForTokenClassification.from_pretrained(
            model_name_or_path, torchscript=True
        )

    def forward(
        self, input_ids, attention_mask
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Forward pass on this NER module requires input_ids and attention_mask.

        * **input_ids** - Tensor generated from FastNN `Processor` class
        * **attention_mask** - Tensor generated from FastNN `Processor` class
        * ** &ast;args ** - args available to make abstraction easier for variant `Processor` classes
        """
        logits = self.model(
            input_ids=input_ids, attention_mask=attention_mask
        )
        return logits[0], input_ids
